from typing import (
    List,
    Optional,
    Union,
)

from eth_typing import (
    Address,
    ChecksumAddress,
    Hash32,
    HexStr,
)
from eth_utils import (
    is_checksum_address,
)
from eth_utils.toolz import (
    assoc,
)

from htdfsdk.web3._utils.compat import (
    Literal,
)
from htdfsdk.web3._utils.personal import (
    ec_recover,
    ecRecover,
    import_raw_key,
    importRawKey,
    list_accounts,
    listAccounts,
    new_account,
    newAccount,
    send_transaction,
    sendTransaction,
    sign,
    sign_typed_data,
    signTypedData,
    unlock_account,
    unlockAccount,
)
from htdfsdk.web3._utils.rpc_abi import (
    RPC,
)
from htdfsdk.web3.module import (
    Module,
    ModuleV2,
)
from htdfsdk.web3.types import (
    ENS,
    BlockIdentifier,
    EnodeURI,
    ParityBlockTrace,
    ParityFilterParams,
    ParityFilterTrace,
    ParityMode,
    ParityNetPeers,
    ParityTraceMode,
    TxParams,
    _Hash32,
)


class ParityPersonal(ModuleV2):
    """
    https://wiki.parity.io/JSONRPC-personal-module
    """
    ec_recover = ec_recover
    import_raw_key = import_raw_key
    list_accounts = list_accounts
    new_account = new_account
    send_transaction = send_transaction
    sign = sign
    sign_typed_data = sign_typed_data
    unlock_account = unlock_account
    # deprecated
    ecRecover = ecRecover
    importRawKey = importRawKey
    listAccounts = listAccounts
    newAccount = newAccount
    sendTransaction = sendTransaction
    signTypedData = signTypedData
    unlockAccount = unlockAccount


class Parity(Module):
    """
    https://paritytech.github.io/wiki/JSONRPC-parity-module
    """
    defaultBlock: Literal["latest"] = "latest"  # noqa: E704
    personal: ParityPersonal

    def enode(self) -> EnodeURI:
        return self.web3.manager.request_blocking(
            RPC.parity_enode,
            [],
        )

    def listStorageKeys(
        self,
        address: Union[Address, ChecksumAddress, ENS],
        quantity: int,
        hash_: Hash32,
        block_identifier: Optional[BlockIdentifier] = None,
    ) -> List[Hash32]:
        if block_identifier is None:
            block_identifier = self.defaultBlock
        return self.web3.manager.request_blocking(
            RPC.parity_listStorageKeys,
            [address, quantity, hash_, block_identifier],
        )

    def netPeers(self) -> ParityNetPeers:
        return self.web3.manager.request_blocking(
            RPC.parity_netPeers,
            [],
        )

    def addReservedPeer(self, url: EnodeURI) -> bool:
        return self.web3.manager.request_blocking(
            RPC.parity_addReservedPeer,
            [url],
        )

    def traceReplayTransaction(
        self, transaction_hash: _Hash32, mode: ParityTraceMode = ['trace']
    ) -> ParityBlockTrace:
        return self.web3.manager.request_blocking(
            RPC.trace_replayTransaction,
            [transaction_hash, mode],
        )

    def traceReplayBlockTransactions(
        self, block_identifier: BlockIdentifier, mode: ParityTraceMode = ['trace']
    ) -> List[ParityBlockTrace]:
        return self.web3.manager.request_blocking(
            RPC.trace_replayBlockTransactions,
            [block_identifier, mode]
        )

    def traceBlock(self, block_identifier: BlockIdentifier) -> List[ParityBlockTrace]:
        return self.web3.manager.request_blocking(
            RPC.trace_block,
            [block_identifier]
        )

    def traceFilter(self, params: ParityFilterParams) -> List[ParityFilterTrace]:
        return self.web3.manager.request_blocking(
            RPC.trace_filter,
            [params]
        )

    def traceTransaction(self, transaction_hash: _Hash32) -> List[ParityFilterTrace]:
        return self.web3.manager.request_blocking(
            RPC.trace_transaction,
            [transaction_hash]
        )

    def traceCall(
        self,
        transaction: TxParams,
        mode: ParityTraceMode = ['trace'],
        block_identifier: Optional[BlockIdentifier] = None
    ) -> ParityBlockTrace:
        # TODO: move to middleware
        if 'from' not in transaction and is_checksum_address(self.web3.eth.defaultAccount):
            transaction = assoc(transaction, 'from', self.web3.eth.defaultAccount)

        # TODO: move to middleware
        if block_identifier is None:
            block_identifier = self.defaultBlock
        return self.web3.manager.request_blocking(
            RPC.trace_call,
            [transaction, mode, block_identifier],
        )

    def traceRawTransaction(
        self, raw_transaction: HexStr, mode: ParityTraceMode = ['trace']
    ) -> ParityBlockTrace:
        return self.web3.manager.request_blocking(
            RPC.trace_rawTransaction,
            [raw_transaction, mode],
        )

    def setMode(self, mode: ParityMode) -> bool:
        return self.web3.manager.request_blocking(
            RPC.parity_setMode,
            [mode]
        )

    def mode(self) -> ParityMode:
        return self.web3.manager.request_blocking(
            RPC.parity_mode,
            []
        )



from typing import (
    NoReturn,
)

from htdfsdk.web3._utils.net import (
    listening,
    peer_count,
    peerCount,
    version,
)
from htdfsdk.web3.module import (
    ModuleV2,
)


class Net(ModuleV2):
    """
        https://github.com/ethereum/wiki/wiki/JSON-RPC
    """

    _listening = listening
    _peer_count = peer_count
    _peerCount = peerCount
    _version = version

    @property
    def listening(self) -> bool:
        return self._listening()

    @property
    def peer_count(self) -> int:
        return self._peer_count()

    @property
    def peerCount(self) -> int:
        return self._peerCount()

    @property
    def chainId(self) -> NoReturn:
        raise DeprecationWarning("This method has been deprecated in EIP 1474.")

    @property
    def version(self) -> str:
        return self._version()



import sys

if sys.version_info < (3, 6):
    raise EnvironmentError(
        "Python 3.6 or above is required. ")


import htdfsdk.web3 as web3

from htdfsdk.htdfsdk import (
    HtdfTxBuilder,
    HtdfRPC,
    Address,
    HtdfPrivateKey,
    ValidatorAddress,
    HtdfDelegateTxBuilder,
    HtdfWithdrawDelegateRewardsTxBuilder,
    HtdfSetUndelegateStatusTxBuilder,
    HtdfUndelegateTxBuilder,
    HtdfEditValidatorInfoTxBuilder
)

from htdfsdk.contract import (
    HtdfContract
)

from eth_utils import (
    to_checksum_address,
    remove_0x_prefix
)


from htdfsdk.utils import *

__all__ = [
    'utils',
    "HtdfRPC",
    "HtdfTxBuilder",
    "Address",
    "HtdfPrivateKey",
    "web3",
    "HtdfContract",
    "to_checksum_address",
    "remove_0x_prefix",
    "ValidatorAddress",
    "HtdfDelegateTxBuilder",
    "HtdfWithdrawDelegateRewardsTxBuilder",
    "HtdfSetUndelegateStatusTxBuilder",
    "HtdfUndelegateTxBuilder",
    "HtdfEditValidatorInfoTxBuilder",
]



# coding:utf8
# author: yqq
# date: 2020/12/15 16:45
# descriptions:
from pprint import pprint
from typing import Dict, List

import json
import hashlib
import time
import ecdsa
# import traceback
import coincurve
import base64
import requests
import logging
from typing import Tuple, Dict
from binascii import hexlify, unhexlify
from bech32 import bech32_decode, convertbits, bech32_encode
# from func_timeout import func_set_timeout
import os

from .utils import htdf_to_satoshi
from .template import *


class Account:
    address: str
    balance_satoshi: int
    account_number: int
    sequence: int


class Address:

    def __init__(self, address: [str, bytes]):
        self.__address = address
        if not Address.is_valid_address(self.__address):
            raise Exception("invalid address: {}".format(self.__address))
        self.__hex_address = Address.bech32_to_hexaddr(self.__address)
        pass

    @classmethod
    def bech32_to_hexaddr(cls, bech32_address: str) -> str:
        assert len(bech32_address) > 0

        hrp, data = bech32_decode(bech32_address)
        hexbytes = convertbits(data, 5, 8, pad=False)
        hexstraddr = ''.join(['%02x' % x for x in hexbytes])
        return hexstraddr.upper()

    @classmethod
    def hexaddr_to_bech32(cls, hex_address: str, hrp: str = 'htdf') -> str:
        assert len(hex_address) > 0 and (len(hex_address) & 1 == 0)
        hexaddr = hex_address[2:] if hex_address[:2] == '0x' else hex_address

        data = [int(x) for x in unhexlify(hexaddr)]
        addr = bech32_encode(hrp, convertbits(data, 8, 5))
        return addr

    @classmethod
    def is_valid_address(cls, address: str) -> bool:
        """ check address validation"""
        if not (len(address) == 43 and address.islower()):
            return False

        prefix, data = bech32_decode(address)
        if prefix is None or data is None:
            return False

        if prefix == 'htdf' and len(data) == 32:
            return True
        return False

    @property
    def address(self) -> str:
        """bech32 encoding address """
        return self.__address

    def bech32_address(self) -> str:
        """
        bech32 encoding address
        :return:
        """
        return self.__address

    # def checksum_hex_address(self):
    #     return

    @property
    def hex_address(self) -> str:
        """
        hex encoding address
        :return:
        """
        return self.__hex_address

    def __str__(self):
        """bech32 encoding address"""
        return self.__address

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.address == other.address




class ValidatorAddress:

    def __init__(self, address: [str, bytes]):
        self.__address = address
        if not self.is_valid_address(self.__address):
            raise Exception("invalid address: {}".format(self.__address))
        self.__hex_address = Address.bech32_to_hexaddr(self.__address)
        pass

    @classmethod
    def bech32_to_hexaddr(cls, bech32_address: str) -> str:
        assert len(bech32_address) > 0

        hrp, data = bech32_decode(bech32_address)
        hexbytes = convertbits(data, 5, 8, pad=False)
        hexstraddr = ''.join(['%02x' % x for x in hexbytes])
        return hexstraddr.upper()

    @classmethod
    def hexaddr_to_bech32(cls, hex_address: str, hrp: str = 'htdfvaloper') -> str:
        assert len(hex_address) > 0 and (len(hex_address) & 1 == 0)
        hexaddr = hex_address[2:] if hex_address[:2] == '0x' else hex_address

        data = [int(x) for x in unhexlify(hexaddr)]
        addr = bech32_encode(hrp, convertbits(data, 8, 5))
        return addr

    @classmethod
    def is_valid_address(cls, address: str) -> bool:
        """ check address validation"""
        if not (len(address) == 50 and address.islower()):
            return False

        prefix, data = bech32_decode(address)
        if prefix is None or data is None:
            return False

        if prefix == 'htdfvaloper' and len(data) == 32:
            return True
        return False

    @property
    def address(self) -> str:
        """bech32 encoding address """
        return self.__address

    def bech32_address(self) -> str:
        """
        bech32 encoding address
        :return:
        """
        return self.__address

    # def checksum_hex_address(self):
    #     return

    @property
    def hex_address(self) -> str:
        """
        hex encoding address
        :return:
        """
        return self.__hex_address

    def __str__(self):
        """bech32 encoding address"""
        return self.__address

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.address == other.address



class HtdfPrivateKey:

    def __init__(self, private_key_hex: [str, bytes]):

        if len(private_key_hex) == 0:
            private_key_hex = self.create_random_private_key()
        else:
            assert len(unhexlify(private_key_hex)) == 32, 'invalid parameter: private_key_hex'

        self.__private_key_hex = private_key_hex
        sk = ecdsa.SigningKey.from_string(unhexlify(self.__private_key_hex), curve=ecdsa.SECP256k1)
        s = sk.get_verifying_key().to_string(encoding='compressed')
        self.__public_key_compressed = hexlify(s).decode('latin1')

        uck = sk.get_verifying_key().to_string(encoding='uncompressed')
        self.__public_key_uncompressed = hexlify(uck).decode('latin1')
        pass

    @classmethod
    def create_random_private_key(cls, salt: str = '') -> str:
        # reference : https://en.bitcoin.it/wiki/Private_key
        MAX_P = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364140
        while True:
            randkey = hexlify(os.urandom(32))
            hash = hexlify(hashlib.sha256(randkey + salt.encode('utf8')).digest())
            if (1 << 32) < int(hash, 16) < MAX_P:
                return hash.decode('latin1')
        pass

    @property
    def private_key(self) -> str:
        return self.__private_key_hex

    @property
    def private_key_bytes(self) -> bytes:
        return unhexlify(self.__private_key_hex)

    def public_key(self, compressed: bool = True) -> str:
        return self.__public_key_compressed if compressed else self.__public_key_uncompressed

    @classmethod
    def pubkey_to_address(cls, public_key_compressed, hrp='htdf'):
        ripemd160 = hashlib.new('ripemd160')
        ripemd160.update(hashlib.sha256(unhexlify(public_key_compressed)).digest())
        r160data = ripemd160.digest()
        data = [r160data[i] for i in range(len(r160data))]
        address = bech32_encode(hrp, convertbits(data, 8, 5))
        return address

    @property
    def address(self) -> Address:
        return Address(self.bech32_address)

    @property
    def bech32_address(self) -> str:
        return self.pubkey_to_address(self.__public_key_compressed)

    @property
    def hex_address(self) -> str:
        return Address.bech32_to_hexaddr(self.bech32_address)

    def sign(self, hash: bytes) -> Tuple:
        pk = coincurve.PrivateKey(self.private_key_bytes)
        signature = pk.sign_recoverable(hash, hasher=None)
        # v = safe_ord(signature[64]) + 27
        r = signature[0:32]
        s = signature[32:64]
        return r, s

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.private_key == other.private_key

    def __str__(self):
        return self.__private_key_hex


class HtdfRPC(object):

    def __init__(self, chaid_id: str, rpc_host: str, rpc_port: int):
        self.chain_id = chaid_id
        self.rpc_host = rpc_host
        self.rpc_port = rpc_port
        self.node_ip_port = '{}:{}'.format(self.rpc_host, self.rpc_port)
        pass

    def get_account_info(self, address: str) -> [Account, None]:
        """
        get account info
        :param address: address
        :return: account info
        """

        url = 'http://{0}/auth/accounts/{1}'.format(self.node_ip_port.strip(), address.strip())
        rsp = requests.get(url)

        if rsp.status_code != 200:
            if rsp.status_code == 204:
                return None
                # raise Exception('not found any account info ')
            raise Exception('get account info error: {}'.format(rsp.status_code))

        rsp = rsp.json()
        acc = Account()
        acc.address = address
        acc.account_number = int(rsp['value']['account_number'])
        acc.sequence = int(rsp['value']['sequence'])
        if rsp['value']['coins'] is not None:
            acc.balance_satoshi = htdf_to_satoshi(float(rsp['value']['coins'][0]['amount']))
        else:
            acc.balance_satoshi = 0
        return acc

    def get_balance(self, address: str) -> int:
        """ get account balance in satoshi"""
        acc = self.get_account_info(address=address)
        if acc:
            return acc.balance_satoshi
        return 0

    def get_block(self, block_number: int) -> [Dict, None]:
        """ get block details by block number"""

        url = 'http://{0}/block_detail/{1}'.format(self.node_ip_port.strip(), block_number)
        rsp = requests.get(url)
        if rsp.status_code != 200:
            return None
        tx = rsp.json()
        return tx

    def get_latest_block(self) -> [Dict, None]:
        """ get latest block details """

        url = 'http://{0}/blocks/latest'.format(self.node_ip_port.strip())
        rsp = requests.get(url)
        if rsp.status_code != 200:
            return None
        tx = rsp.json()
        return tx

    def get_mempool_transaction(self, transaction_hash: str) -> [Dict, None]:
        """
        search a tx in mempool by tx_hash, if it not in mempool return None .
        NOTE: a confirmed(successed or failed) transaction couldn't be found in mempool
        """
        req_url = 'http://{0}/mempool/txs/{1}'.format(self.node_ip_port.strip(), transaction_hash.strip())
        rsp = requests.get(url=req_url)
        if rsp.status_code == 400:
            return None
        tx = rsp.json()
        return tx

    def get_mempool_trasactions(self) -> [List[Dict], None]:
        """ get the frontly 101 transactions """
        # raise NotImplementedError
        req_url = 'http://{0}/mempool/txs'.format(self.node_ip_port.strip())
        rsp = requests.get(url=req_url)
        if rsp.status_code != 200:
            return None
        tx = rsp.json()
        return tx

    def get_mempool_transaction_count(self) -> [Dict, None]:
        """ get mempool transaction count """
        # raise NotImplementedError
        req_url = 'http://{0}/mempool/txscount'.format(self.node_ip_port.strip())
        rsp = requests.get(url=req_url)
        if rsp.status_code != 200:
            return None
        tx = rsp.json()
        return tx

    # @func_set_timeout(30)
    def get_transaction_until_timeout(self, transaction_hash: str, timeout_secs: float = 15,
                                     interval_secs: float = 5) -> [Dict, None]:
        assert 0.0 < interval_secs < timeout_secs, 'interval_secs must less than timeout secs'
        start_time = time.time()
        while True:
            url = 'http://{0}/txs/{1}'.format(self.node_ip_port, transaction_hash.strip())
            rsp = requests.get(url)
            if rsp.status_code == 404:
                cur_time = time.time()
                if cur_time - start_time >= timeout_secs:
                    return None
                if timeout_secs - (cur_time - start_time) >= interval_secs:
                    time.sleep(interval_secs)
                else:
                    time.sleep(timeout_secs - (cur_time - start_time))
                continue
            tx = rsp.json()
            return tx
        pass

    def get_transaction_receipt_until_timeout(self, transaction_hash: str, timeout_secs: float = 15,
                                      interval_secs: float = 5) -> [Dict, None]:
        assert 0.0 < interval_secs < timeout_secs, 'interval_secs must less than timeout secs'
        start_time = time.time()
        while True:
            url = 'http://{0}/v2/tx-receipt/{1}'.format(self.node_ip_port, transaction_hash.strip())
            rsp = requests.get(url)
            if rsp.status_code == 404:
                cur_time = time.time()
                if cur_time - start_time >= timeout_secs:
                    return None
                if timeout_secs - (cur_time - start_time) >= interval_secs:
                    time.sleep(interval_secs)
                else:
                    time.sleep(timeout_secs - (cur_time - start_time))
                continue
            tx = rsp.json()
            return tx
        pass

    def get_transaction_receipt(self, transaction_hash: str) -> [Dict, None]:
        """
        get transaction receipt by hash
        :param transaction_hash:  tx hash
        :return: if found , return transaction details , else return None
        """

        url = 'http://{0}/v2/tx-receipt/{1}'.format(self.node_ip_port.strip(), transaction_hash.strip())
        rsp = requests.get(url)
        if rsp.status_code == 404:
            return None
        tx = rsp.json()
        return tx

    def get_transaction(self, transaction_hash: str) -> [Dict, None]:
        """
        get transaction by hash
        :param transaction_hash:  tx hash
        :return: if found , return transaction details , else return None
        """

        url = 'http://{0}/txs/{1}'.format(self.node_ip_port.strip(), transaction_hash.strip())
        rsp = requests.get(url)
        if rsp.status_code == 404:
            return None
        tx = rsp.json()
        return tx

    def broadcast_tx(self, tx_hex: str) -> str:
        """
        broadcast transaction
        :param tx_hex:  transaction is hex-string
        :return:  transaction hash
        """

        # check tx_hex
        unhexlify(tx_hex)

        post_data = {'tx': tx_hex}
        post_data = json.dumps(post_data)
        url = 'http://%s/hs/broadcast' % (self.node_ip_port)
        rsp = requests.post(url, post_data)

        if rsp.status_code != 200:
            raise Exception(rsp.text)

        rsp_json = rsp.json()
        if 'code' in rsp_json or 'raw_log' in rsp_json:
            raise Exception(rsp.text)

        txhash = str(rsp_json['txhash'])

        return txhash

    def get_upgrade_info(self) -> Dict:
        """
        get upgrade info
        :return:
        """
        url = 'http://{0}/upgrade_info'.format(self.node_ip_port.strip())
        rsp = requests.get(url)
        if rsp.status_code != 200:
            raise Exception("get upgrade info error:{}".format(rsp.text))
        tx = rsp.json()
        return tx

    def contract_call(self, contract_address: str, hex_data: str) -> str:
        url = 'http://{0}/hs/contract/{1}/{2}'.format(self.node_ip_port.strip(), contract_address, hex_data)
        rsp = requests.get(url)
        if rsp.status_code != 200:
            raise Exception("contract_call:{}".format(rsp.text))
        rsp = rsp.text
        rsp = rsp.replace('"', '')
        return rsp

    def __str__(self):
        pass


class HtdfTxBuilder(object):

    def __init__(self, from_address: Address, to_address: [Address, None], amount_satoshi: int,
                 sequence: int, account_number: int, chain_id: str,
                 gas_price: int = 100, gas_wanted: int = 30000, data: str = '', memo: str = ''):
        self.from_address = from_address
        self.to_address = '' if to_address is None else to_address
        self.amount_satoshi = amount_satoshi
        self.data = data
        self.memo = memo
        self.sequence = sequence
        self.account_number = account_number
        self.chain_id = chain_id
        self.gas_price = gas_price
        self.gas_wanted = gas_wanted
        pass

    def build_and_sign(self, private_key: HtdfPrivateKey) -> str:
        """
        build and sign transaction
        :param private_key:
        :return: signed transaction
        """

        # logging.info('account_number : {}, sequence: {} '.format(account_number, sequence))

        # step 3 : format raw transaction
        fmt_unsigned_txstr = UNSIGNED_TX_TEMPLATE.replace(' ', '').replace('\t', '').replace('\n', '')
        fmt_unsigned_txstr = fmt_unsigned_txstr % (
            self.account_number, self.chain_id, self.gas_price, self.gas_wanted, self.memo,
            self.amount_satoshi, self.data, self.from_address, self.gas_price, self.gas_wanted,
            self.to_address, self.sequence)

        # logging.info("formatted raw transaction str: {}".format(fmt_unsigned_txstr))

        # step 4 : make signature
        shadata = hashlib.sha256(fmt_unsigned_txstr.encode('utf-8')).digest()
        logging.info("sha256(fmt_unsigned_txstr): {}".format(hexlify(shadata)))

        r, s = private_key.sign(shadata)
        # logging.info('r:' + hexlify(r).decode(encoding='utf8'))
        # logging.info('s:' + hexlify(s).decode(encoding='utf8'))

        b64sig = base64.b64encode(r + s).decode(encoding='utf8')
        # logging.info("base64encode(signature) : {}".format(b64sig))

        # step 5 : format broadcast string
        pubkey = private_key.public_key()  # self.__privkey_to_pubkey(privkey=private_key)
        b64pubkey = base64.b64encode(unhexlify(pubkey)).decode(encoding='utf8')
        # logging.info("base64encode(public key) :" + b64pubkey)

        fmt_broadcast_str = BROADCAST_TX_TEMPLATE.replace(' ', '').replace('\n', '').replace('\t', '')
        fmt_broadcast_str = fmt_broadcast_str % (
            self.from_address, self.to_address, self.amount_satoshi,
            self.data, self.gas_price, self.gas_wanted, self.gas_wanted, self.gas_price, b64pubkey, b64sig, self.memo)
        # logging.info("broadcast str: {}".format(fmt_broadcast_str))

        broadcast_data = hexlify(bytes(fmt_broadcast_str, encoding='utf8')).decode(encoding='utf8')
        return broadcast_data

    def __str__(self):
        pass




class HtdfDelegateTxBuilder(object):

    def __init__(self, delegator_address: Address, validator_address: ValidatorAddress, amount_satoshi: int,
                 sequence: int, account_number: int, chain_id: str,
                 gas_price: int = 100, gas_wanted: int = 30000,  memo: str = ''):
        self.delegator_address = delegator_address
        self.validator_address = validator_address
        self.amount_satoshi = amount_satoshi
        self.memo = memo
        self.sequence = sequence
        self.account_number = account_number
        self.chain_id = chain_id
        self.gas_price = gas_price
        self.gas_wanted = gas_wanted
        pass

    def build_and_sign(self, private_key: HtdfPrivateKey) -> str:
        """
        build and sign transaction
        :param private_key:
        :return: signed transaction
        """

        # logging.info('account_number : {}, sequence: {} '.format(account_number, sequence))

        # step 3 : format raw transaction
        fmt_unsigned_txstr = UNSIGNED_DELEGATE_TX_TEMPLATE.replace(' ', '').replace('\t', '').replace('\n', '')
        fmt_unsigned_txstr = fmt_unsigned_txstr % (
            self.account_number, self.chain_id, self.gas_price, self.gas_wanted, self.memo,
            self.amount_satoshi, self.delegator_address, self.validator_address,  self.sequence)

        # logging.info("formatted raw transaction str: {}".format(fmt_unsigned_txstr))

        # step 4 : make signature
        shadata = hashlib.sha256(fmt_unsigned_txstr.encode('utf-8')).digest()
        logging.info("sha256(fmt_unsigned_txstr): {}".format(hexlify(shadata)))

        r, s = private_key.sign(shadata)
        # logging.info('r:' + hexlify(r).decode(encoding='utf8'))
        # logging.info('s:' + hexlify(s).decode(encoding='utf8'))

        b64sig = base64.b64encode(r + s).decode(encoding='utf8')
        # logging.info("base64encode(signature) : {}".format(b64sig))

        # step 5 : format broadcast string
        pubkey = private_key.public_key()  # self.__privkey_to_pubkey(privkey=private_key)
        b64pubkey = base64.b64encode(unhexlify(pubkey)).decode(encoding='utf8')
        # logging.info("base64encode(public key) :" + b64pubkey)

        fmt_broadcast_str = BROADCAST_DELEGATE_TX_TEMPLATE.replace(' ', '').replace('\n', '').replace('\t', '')
        fmt_broadcast_str = fmt_broadcast_str % (
            self.delegator_address, self.validator_address, self.amount_satoshi,
            self.gas_wanted, self.gas_price, b64pubkey, b64sig, self.memo)
        # logging.info("broadcast str: {}".format(fmt_broadcast_str))
        print("broadcast str: {}".format(fmt_broadcast_str))
        btx = json.loads(fmt_broadcast_str)
        pprint(btx)

        broadcast_data = hexlify(bytes(fmt_broadcast_str, encoding='utf8')).decode(encoding='utf8')
        return broadcast_data

    def __str__(self):
        pass



class HtdfSetUndelegateStatusTxBuilder(object):

    def __init__(self, delegator_address: Address, validator_address: ValidatorAddress, status: bool,
                 sequence: int, account_number: int, chain_id: str,
                 gas_price: int = 100, gas_wanted: int = 30000,  memo: str = ''):
        self.delegator_address = delegator_address
        self.validator_address = validator_address
        self.status = 'true' if status else 'false'
        self.memo = memo
        self.sequence = sequence
        self.account_number = account_number
        self.chain_id = chain_id
        self.gas_price = gas_price
        self.gas_wanted = gas_wanted
        pass


    def build_and_sign(self, private_key: HtdfPrivateKey) -> str:
        """
        build and sign transaction
        :param private_key:
        :return: signed transaction
        """

        # logging.info('account_number : {}, sequence: {} '.format(account_number, sequence))

        # step 3 : format raw transaction
        fmt_unsigned_txstr = UNSIGNED_SET_UNDELEGATE_STATUS_TX_TEMPLATE.replace(' ', '').replace('\t', '').replace('\n', '')
        fmt_unsigned_txstr = fmt_unsigned_txstr % (
            self.account_number, self.chain_id, self.gas_price, self.gas_wanted, self.memo,
            self.status, self.delegator_address, self.validator_address,  self.sequence)

        # logging.info("formatted raw transaction str: {}".format(fmt_unsigned_txstr))

        # step 4 : make signature
        shadata = hashlib.sha256(fmt_unsigned_txstr.encode('utf-8')).digest()
        logging.info("sha256(fmt_unsigned_txstr): {}".format(hexlify(shadata)))

        r, s = private_key.sign(shadata)
        # logging.info('r:' + hexlify(r).decode(encoding='utf8'))
        # logging.info('s:' + hexlify(s).decode(encoding='utf8'))

        b64sig = base64.b64encode(r + s).decode(encoding='utf8')
        # logging.info("base64encode(signature) : {}".format(b64sig))

        # step 5 : format broadcast string
        pubkey = private_key.public_key()  # self.__privkey_to_pubkey(privkey=private_key)
        b64pubkey = base64.b64encode(unhexlify(pubkey)).decode(encoding='utf8')
        # logging.info("base64encode(public key) :" + b64pubkey)

        fmt_broadcast_str = BROADCAST_SET_UNDELEGATE_STATUS_TX_TEMPLATE.replace(' ', '').replace('\n', '').replace('\t', '')
        fmt_broadcast_str = fmt_broadcast_str % (
            self.status, self.delegator_address, self.validator_address,
            self.gas_wanted, self.gas_price, b64pubkey, b64sig, self.memo)
        # logging.info("broadcast str: {}".format(fmt_broadcast_str))
        print("broadcast str: {}".format(fmt_broadcast_str))
        btx = json.loads(fmt_broadcast_str)
        pprint(btx)

        broadcast_data = hexlify(bytes(fmt_broadcast_str, encoding='utf8')).decode(encoding='utf8')
        return broadcast_data

    def __str__(self):
        pass


class HtdfUndelegateTxBuilder(object):

    def __init__(self, delegator_address: Address, validator_address: ValidatorAddress, amount_satoshi: int,
                 sequence: int, account_number: int, chain_id: str,
                 gas_price: int = 100, gas_wanted: int = 30000, memo: str = ''):
        self.delegator_address = delegator_address
        self.validator_address = validator_address
        self.amount_satoshi = amount_satoshi
        self.memo = memo
        self.sequence = sequence
        self.account_number = account_number
        self.chain_id = chain_id
        self.gas_price = gas_price
        self.gas_wanted = gas_wanted
        pass

    def build_and_sign(self, private_key: HtdfPrivateKey) -> str:
        """
        build and sign transaction
        :param private_key:
        :return: signed transaction
        """

        # logging.info('account_number : {}, sequence: {} '.format(account_number, sequence))

        # step 3 : format raw transaction
        fmt_unsigned_txstr = UNSIGNED_UNDELEGATE_TX_TEMPLATE.replace(' ', '').replace('\t', '').replace('\n',
                                                                                                                   '')
        fmt_unsigned_txstr = fmt_unsigned_txstr % (
            self.account_number, self.chain_id, self.gas_price, self.gas_wanted, self.memo,
            self.amount_satoshi, self.delegator_address, self.validator_address, self.sequence)

        # logging.info("formatted raw transaction str: {}".format(fmt_unsigned_txstr))

        # step 4 : make signature
        shadata = hashlib.sha256(fmt_unsigned_txstr.encode('utf-8')).digest()
        logging.info("sha256(fmt_unsigned_txstr): {}".format(hexlify(shadata)))

        r, s = private_key.sign(shadata)
        # logging.info('r:' + hexlify(r).decode(encoding='utf8'))
        # logging.info('s:' + hexlify(s).decode(encoding='utf8'))

        b64sig = base64.b64encode(r + s).decode(encoding='utf8')
        # logging.info("base64encode(signature) : {}".format(b64sig))

        # step 5 : format broadcast string
        pubkey = private_key.public_key()  # self.__privkey_to_pubkey(privkey=private_key)
        b64pubkey = base64.b64encode(unhexlify(pubkey)).decode(encoding='utf8')
        # logging.info("base64encode(public key) :" + b64pubkey)

        fmt_broadcast_str = BROADCAST_UNDELEGATE_TX_TEMPLATE.replace(' ', '').replace('\n', '').replace('\t', '')
        fmt_broadcast_str = fmt_broadcast_str % (
            self.delegator_address, self.validator_address, self.amount_satoshi,
            self.gas_wanted, self.gas_price, b64pubkey, b64sig, self.memo)
        # logging.info("broadcast str: {}".format(fmt_broadcast_str))
        print("broadcast str: {}".format(fmt_broadcast_str))
        btx = json.loads(fmt_broadcast_str)
        pprint(btx)

        broadcast_data = hexlify(bytes(fmt_broadcast_str, encoding='utf8')).decode(encoding='utf8')
        return broadcast_data

    def __str__(self):
        pass





class HtdfWithdrawDelegateRewardsTxBuilder(object):

    def __init__(self, delegator_address: Address, validator_address: ValidatorAddress,
                 sequence: int, account_number: int, chain_id: str,
                 gas_price: int = 100, gas_wanted: int = 30000, memo: str = ''):
        self.delegator_address = delegator_address
        self.validator_address = validator_address
        self.memo = memo
        self.sequence = sequence
        self.account_number = account_number
        self.chain_id = chain_id
        self.gas_price = gas_price
        self.gas_wanted = gas_wanted
        pass

    def build_and_sign(self, private_key: HtdfPrivateKey) -> str:
        """
        build and sign transaction
        :param private_key:
        :return: signed transaction
        """

        # logging.info('account_number : {}, sequence: {} '.format(account_number, sequence))

        # step 3 : format raw transaction
        fmt_unsigned_txstr = UNSIGNED_WITHDRAW_DELEGATE_REWARDS_TX_TEMPLATE.replace(' ', '').replace('\t', '').replace('\n',
                                                                                                                   '')
        fmt_unsigned_txstr = fmt_unsigned_txstr % (
            self.account_number, self.chain_id, self.gas_price, self.gas_wanted, self.memo,
             self.delegator_address, self.validator_address, self.sequence)

        # logging.info("formatted raw transaction str: {}".format(fmt_unsigned_txstr))

        # step 4 : make signature
        shadata = hashlib.sha256(fmt_unsigned_txstr.encode('utf-8')).digest()
        logging.info("sha256(fmt_unsigned_txstr): {}".format(hexlify(shadata)))

        r, s = private_key.sign(shadata)
        # logging.info('r:' + hexlify(r).decode(encoding='utf8'))
        # logging.info('s:' + hexlify(s).decode(encoding='utf8'))

        b64sig = base64.b64encode(r + s).decode(encoding='utf8')
        # logging.info("base64encode(signature) : {}".format(b64sig))

        # step 5 : format broadcast string
        pubkey = private_key.public_key()  # self.__privkey_to_pubkey(privkey=private_key)
        b64pubkey = base64.b64encode(unhexlify(pubkey)).decode(encoding='utf8')
        # logging.info("base64encode(public key) :" + b64pubkey)

        fmt_broadcast_str = BROADCAST_WITHDRAW_DELEGATE_REWARDS_TX_TEMPLATE.replace(' ', '').replace('\n', '').replace('\t', '')
        fmt_broadcast_str = fmt_broadcast_str % (
            self.delegator_address, self.validator_address,
            self.gas_wanted, self.gas_price, b64pubkey, b64sig, self.memo)
        # logging.info("broadcast str: {}".format(fmt_broadcast_str))
        print("broadcast str: {}".format(fmt_broadcast_str))
        btx = json.loads(fmt_broadcast_str)
        pprint(btx)

        broadcast_data = hexlify(bytes(fmt_broadcast_str, encoding='utf8')).decode(encoding='utf8')
        return broadcast_data

    def __str__(self):
        pass





class HtdfEditValidatorInfoTxBuilder(object):

    def __init__(self,  validator_address: ValidatorAddress, commission_rate: str,
                 sequence: int, account_number: int, chain_id: str,
                 gas_price: int = 100, gas_wanted: int = 30000, memo: str = '',
                 details: str = '', identity: str = '', moniker: str = '', website: str='',
                 min_self_delegation: str = '1'):
        self.validator_address = validator_address
        self.memo = memo
        self.sequence = sequence
        self.account_number = account_number
        self.chain_id = chain_id
        self.gas_price = gas_price
        self.gas_wanted = gas_wanted
        self.details = details
        self.identity = identity
        self.moniker = moniker
        self.website = website
        self.min_self_delegation = min_self_delegation

        # precision is 18
        self.commission_rate = '{:0<20}'.format(commission_rate)

        pass

    def build_and_sign(self, private_key: HtdfPrivateKey) -> str:
        """
        build and sign transaction
        :param private_key:
        :return: signed transaction
        """

        # logging.info('account_number : {}, sequence: {} '.format(account_number, sequence))

        # step 3 : format raw transaction
        fmt_unsigned_txstr = UNSIGNED_EDIT_VALIDATOR_INFO_TX_TEMPLATE.replace(' ', '').replace('\t', '').replace('\n',
                                                                                                                   '')
        fmt_unsigned_txstr = fmt_unsigned_txstr % (
            self.account_number, self.chain_id, self.gas_price, self.gas_wanted, self.memo,
            self.details, self.identity, self.moniker, self.website, self.validator_address,
            self.commission_rate, self.min_self_delegation, self.sequence)

        print(fmt_unsigned_txstr)

        # untx = json.loads(fmt_unsigned_txstr)
        # pprint(untx)

        # logging.info("formatted raw transaction str: {}".format(fmt_unsigned_txstr))

        # step 4 : make signature
        shadata = hashlib.sha256(fmt_unsigned_txstr.encode('utf-8')).digest()
        logging.info("sha256(fmt_unsigned_txstr): {}".format(hexlify(shadata)))

        r, s = private_key.sign(shadata)
        # logging.info('r:' + hexlify(r).decode(encoding='utf8'))
        # logging.info('s:' + hexlify(s).decode(encoding='utf8'))

        b64sig = base64.b64encode(r + s).decode(encoding='utf8')
        # logging.info("base64encode(signature) : {}".format(b64sig))

        # step 5 : format broadcast string
        pubkey = private_key.public_key()  # self.__privkey_to_pubkey(privkey=private_key)
        b64pubkey = base64.b64encode(unhexlify(pubkey)).decode(encoding='utf8')
        # logging.info("base64encode(public key) :" + b64pubkey)

        fmt_broadcast_str = BROADCAST_EDIT_VALIDATOR_INFO_TX_TEMPLATE.replace(' ', '').replace('\n', '').replace('\t', '')
        fmt_broadcast_str = fmt_broadcast_str % (
            self.details, self.identity, self.moniker, self.website, self.validator_address,
            self.commission_rate, self.min_self_delegation,
            self.gas_wanted, self.gas_price, b64pubkey, b64sig, self.memo)
        # logging.info("broadcast str: {}".format(fmt_broadcast_str))
        print("broadcast str: {}".format(fmt_broadcast_str))
        btx = json.loads(fmt_broadcast_str)
        pprint(btx)

        broadcast_data = hexlify(bytes(fmt_broadcast_str, encoding='utf8')).decode(encoding='utf8')
        return broadcast_data

    def __str__(self):
        pass


from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
)

from eth_utils.toolz import (
    assoc,
)

from htdfsdk.web3.types import (
    RPCEndpoint,
    RPCResponse,
)

if TYPE_CHECKING:
    from htdfsdk.web3 import Web3  # noqa: F401


def gas_price_strategy_middleware(
    make_request: Callable[[RPCEndpoint, Any], Any], web3: "Web3"
) -> Callable[[RPCEndpoint, Any], RPCResponse]:
    """
    Includes a gas price using the gas price strategy
    """
    def middleware(method: RPCEndpoint, params: Any) -> RPCResponse:
        if method == 'eth_sendTransaction':
            transaction = params[0]
            if 'gasPrice' not in transaction:
                generated_gas_price = web3.eth.generateGasPrice(transaction)
                if generated_gas_price is not None:
                    transaction = assoc(transaction, 'gasPrice', generated_gas_price)
                    return make_request(method, [transaction])
        return make_request(method, params)
    return middleware



#coding:utf8
#author: yqq
#date: 2020/12/15 下午5:38
#descriptions:

from decimal import Decimal, getcontext

# getcontext()

def htdf_to_satoshi(amount_htdf: [float, int, str]) -> int:
    return int(Decimal(str(amount_htdf)) * (10 ** 8))




if __name__ == '__main__':
    assert htdf_to_satoshi(139623.71827296) == 13962371827296
    assert htdf_to_satoshi('139623.71827296') == 13962371827296
    assert htdf_to_satoshi(13962371827296) == 13962371827296 * 10 ** 8
    pass




import setuptools


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='htdfsdk',
    version='1.0.1',
    description='htdf python sdk',
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords='htdf sdk',
    py_modules=['htdfsdk'],
    packages=setuptools.find_packages(exclude=["tests"]),
    author='youngqqcn',
    author_email='youngqqcn@gmail.com',
    url='https://github.com/youngqqcn/htdfsdk',
    classifiers=[
        # 'Development Status :: 1 - Planning',
        # 'Development Status :: 2 - Pre-Alpha',
        # 'Development Status :: 3 - Alpha',
        'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable'
        # 'Development Status :: 5 - Production/Stable',  # 当前开发进度等级（测试版，正式版等）
        'Intended Audience :: Developers',  # 模块适用人群
        'Topic :: Software Development :: Code Generators',  # 给模块加话题标签
        'License :: OSI Approved :: MIT License',  # 模块的license

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    python_requires='>=3.6',
    install_requires=[
        'ecdsa',
        'coincurve',
        'bech32',
        'requests',

        "eth-abi>=2.0.0b6,<3.0.0",
        "eth-account>=0.5.3,<0.6.0",
        "eth-hash[pycryptodome]>=0.2.0,<1.0.0",
        "eth-typing>=2.0.0,<3.0.0",
        "eth-utils>=1.9.5,<2.0.0",
        "hexbytes>=0.1.0,<1.0.0",
        "ipfshttpclient==0.7.0a1",
        "jsonschema>=3.2.0,<4.0.0",
        "lru-dict>=1.1.6,<2.0.0",
        "protobuf>=3.10.0,<4",
        "pypiwin32>=223;platform_system=='Windows'",
        "requests>=2.16.0,<3.0.0",
        # remove typing_extensions after python_requires>=3.8, see web3._utils.compat
         "typing-extensions>=3.7.4.1,<4;python_version<'3.8'",
        "websockets>=8.1.0,<9.0.0",
    ],

)


from typing import (
    Any,
    Dict,
    Sequence,
    TypeVar,
)

from htdfsdk.web3.exceptions import (
    ValidationError,
)

T = TypeVar("T")


def attach_modules(parent_module: T, module_definitions: Dict[str, Sequence[Any]]) -> None:
    for module_name, module_info in module_definitions.items():
        module_class = module_info[0]
        module_class.attach(parent_module, module_name)

        if len(module_info) == 2:
            submodule_definitions = module_info[1]
            module = getattr(parent_module, module_name)
            attach_modules(module, submodule_definitions)
        elif len(module_info) != 1:
            raise ValidationError("Module definitions can only have 1 or 2 elements.")



from typing import (
    Callable,
)

from htdfsdk.web3._utils.rpc_abi import (
    RPC,
)
from htdfsdk.web3.method import (
    DeprecatedMethod,
    Method,
    default_root_munger,
)

listening: Method[Callable[[], bool]] = Method(
    RPC.net_listening,
    mungers=[default_root_munger],
)

peer_count: Method[Callable[[], int]] = Method(
    RPC.net_peerCount,
    mungers=[default_root_munger],
)

version: Method[Callable[[], str]] = Method(
    RPC.net_version,
    mungers=[default_root_munger],
)


#
# Deprecated Methods
#
peerCount = DeprecatedMethod(peer_count, 'peerCount', 'peer_count')



#coding:utf8
#author: yqq
#date: 2020/12/15 下午7:16
#descriptions:

UNSIGNED_TX_TEMPLATE = """{\
           "account_number": "%d",\
       	"chain_id": "%s",\
       	"fee": {\
       			"gas_price": "%d",\
       			"gas_wanted": "%d"\
       	},\
           "memo": "%s",\
       	"msgs": [{\
       		"Amount": [{\
       			"amount": "%d",\
                   "denom": "satoshi"\
       		}],\
               "Data": "%s",\
               "From": "%s",\
               "GasPrice": %s,\
               "GasWanted": %s,\
       		"To": "%s"\
       	}],\
           "sequence": "%d"\
           }"""

BROADCAST_TX_TEMPLATE  = """{
           "type": "auth/StdTx",
           "value":{
               "msg": [{
                   "type": "htdfservice/send",  
                   "value":{
                       "From": "%s",
                       "To": "%s",
                       "Amount": [{
                           "denom": "satoshi",
                           "amount": "%d"
                       }],
                       "Data": "%s",
                       "GasPrice": "%d",
                       "GasWanted": "%d"
                   }
               }],
               "fee": {
                   "gas_wanted": "%d",
                   "gas_price": "%d"
               },
               "signatures": [{
                   "pub_key": {
                       "type": "tendermint/PubKeySecp256k1",
                       "value": "%s"
                   },
                   "signature": "%s"
               }],
               "memo": "%s"
           }
       }"""




UNSIGNED_DELEGATE_TX_TEMPLATE = """{\
        "account_number": "%d",\
       	"chain_id": "%s",\
       	"fee": {\
       			"gas_price": "%d",\
       			"gas_wanted": "%d"\
       	},\
        "memo": "%s",\
       	"msgs": [{\
       	    "type":"htdf/MsgDelegate",\
       	    "value":{\
       	        "amount": {\
       			    "amount": "%d",\
                    "denom": "satoshi"\
       		    },\
                "delegator_address": "%s",\
       		    "validator_address": "%s"\
       		}\
       	}],\
        "sequence": "%d"\
        }"""



BROADCAST_DELEGATE_TX_TEMPLATE  = """{
           "type": "auth/StdTx",
           "value":{
               "msg": [{
                   "type": "htdf/MsgDelegate",  
                   "value":{
                       "delegator_address": "%s",
                       "validator_address": "%s",
                       "amount": {
                           "denom": "satoshi",
                           "amount": "%d"
                       }
                   }
               }],
               "fee": {
                   "gas_wanted": "%d",
                   "gas_price": "%d"
               },
               "signatures": [{
                   "pub_key": {
                       "type": "tendermint/PubKeySecp256k1",
                       "value": "%s"
                   },
                   "signature": "%s"
               }],
               "memo": "%s"
           }
       }"""





UNSIGNED_WITHDRAW_DELEGATE_REWARDS_TX_TEMPLATE = """{\
        "account_number": "%d",\
       	"chain_id": "%s",\
       	"fee": {\
       			"gas_price": "%d",\
       			"gas_wanted": "%d"\
       	},\
        "memo": "%s",\
       	"msgs": [{\
       	    "type":"htdf/MsgWithdrawDelegationReward",\
       	    "value":{\
                "delegator_address": "%s",\
       		    "validator_address": "%s"\
       		}\
       	}],\
        "sequence": "%d"\
        }"""



BROADCAST_WITHDRAW_DELEGATE_REWARDS_TX_TEMPLATE  = """{
           "type": "auth/StdTx",
           "value":{
               "msg": [{
                   "type": "htdf/MsgWithdrawDelegationReward",  
                   "value":{
                       "delegator_address": "%s",
                       "validator_address": "%s"
                   }
               }],
               "fee": {
                   "gas_wanted": "%d",
                   "gas_price": "%d"
               },
               "signatures": [{
                   "pub_key": {
                       "type": "tendermint/PubKeySecp256k1",
                       "value": "%s"
                   },
                   "signature": "%s"
               }],
               "memo": "%s"
           }
       }"""



UNSIGNED_UNDELEGATE_TX_TEMPLATE = """{\
        "account_number": "%d",\
       	"chain_id": "%s",\
       	"fee": {\
       			"gas_price": "%d",\
       			"gas_wanted": "%d"\
       	},\
        "memo": "%s",\
       	"msgs": [{\
       	    "type":"htdf/MsgUndelegate",\
       	    "value":{\
       	        "amount": {\
       			    "amount": "%d",\
                    "denom": "satoshi"\
       		    },\
                "delegator_address": "%s",\
       		    "validator_address": "%s"\
       		}\
       	}],\
        "sequence": "%d"\
        }"""



BROADCAST_UNDELEGATE_TX_TEMPLATE  = """{
           "type": "auth/StdTx",
           "value":{
               "msg": [{
                   "type": "htdf/MsgUndelegate",  
                   "value":{
                       "delegator_address": "%s",
                       "validator_address": "%s",
                       "amount": {
                           "denom": "satoshi",
                           "amount": "%d"
                       }
                   }
               }],
               "fee": {
                   "gas_wanted": "%d",
                   "gas_price": "%d"
               },
               "signatures": [{
                   "pub_key": {
                       "type": "tendermint/PubKeySecp256k1",
                       "value": "%s"
                   },
                   "signature": "%s"
               }],
               "memo": "%s"
           }
       }"""





UNSIGNED_SET_UNDELEGATE_STATUS_TX_TEMPLATE = """{\
        "account_number": "%d",\
       	"chain_id": "%s",\
       	"fee": {\
       			"gas_price": "%d",\
       			"gas_wanted": "%d"\
       	},\
        "memo": "%s",\
       	"msgs": [{\
       	    "type":"htdf/MsgSetUndelegateStatus",\
       	    "value":{\
       	        "Status":%s,\
                "delegator_address": "%s",\
       		    "validator_address": "%s"\
       		}\
       	}],\
        "sequence": "%d"\
        }"""



BROADCAST_SET_UNDELEGATE_STATUS_TX_TEMPLATE  = """{
           "type": "auth/StdTx",
           "value":{
               "msg": [{
                   "type": "htdf/MsgSetUndelegateStatus",  
                   "value":{
                        "Status":%s,
                       "delegator_address": "%s",
                       "validator_address": "%s"
                   }
               }],
               "fee": {
                   "gas_wanted": "%d",
                   "gas_price": "%d"
               },
               "signatures": [{
                   "pub_key": {
                       "type": "tendermint/PubKeySecp256k1",
                       "value": "%s"
                   },
                   "signature": "%s"
               }],
               "memo": "%s"
           }
       }"""



UNSIGNED_EDIT_VALIDATOR_INFO_TX_TEMPLATE = """{\
        "account_number": "%d",\
       	"chain_id": "%s",\
       	"fee": {\
       			"gas_price": "%d",\
       			"gas_wanted": "%d"\
       	},\
        "memo": "%s",\
       	"msgs": [{\
       	    "type":"htdf/MsgEditValidator",\
       	    "value":{\
       	        "Description":{\
       	            "details":"%s",\
       	            "identity":"%s",\
       	            "moniker":"%s",\
       	            "website":"%s"\
       	        },\
       	        "address":"%s",\
       	        "commission_rate":"%s",\
       	        "min_self_delegation":"%s"\
       		}\
       	}],\
        "sequence": "%d"\
        }"""



BROADCAST_EDIT_VALIDATOR_INFO_TX_TEMPLATE  = """{
           "type": "auth/StdTx",
           "value":{
               "msg": [{
                   "type": "htdf/MsgEditValidator",  
                   "value":{
                        "Description":{
       	                    "details":"%s",
       	                    "identity":"%s",
       	                    "moniker":"%s",
       	                    "website":"%s"
       	                },
       	                "address":"%s",
       	                "commission_rate":"%s",
       	                "min_self_delegation":"%s"
                   }
               }],
               "fee": {
                   "gas_wanted": "%d",
                   "gas_price": "%d"
               },
               "signatures": [{
                   "pub_key": {
                       "type": "tendermint/PubKeySecp256k1",
                       "value": "%s"
                   },
                   "signature": "%s"
               }],
               "memo": "%s"
           }
       }"""


# coding:utf8
# author: yqq
# date: 2020/12/18 10:44
# descriptions:  contract wrapper for htdf contract transactions
import itertools
import json
from typing import Optional, Any

from eth_abi.exceptions import DecodingError
from eth_typing import HexStr
from eth_utils import to_checksum_address, remove_0x_prefix, decode_hex

from htdfsdk import Address, HtdfRPC
from htdfsdk.web3 import Web3
from htdfsdk.web3._utils.abi import get_abi_output_types, map_abi_data
from htdfsdk.web3._utils.contracts import prepare_transaction, find_matching_fn_abi
from htdfsdk.web3._utils.normalizers import BASE_RETURN_NORMALIZERS
from htdfsdk.web3.contract import Contract, ContractFunction, ACCEPTABLE_EMPTY_STRINGS
from htdfsdk.web3.exceptions import BadFunctionCallOutput
from htdfsdk.web3.types import TxParams


class HtdfContract:

    def __init__(self, rpc: HtdfRPC, address: [Address, None], **kwargs: Any):
        assert address is None or isinstance(address, Address)
        self.rpc = rpc
        self.web3 = Web3()
        self.contract_factory = Contract.factory(web3=self.web3, **kwargs)
        if address is not None :
            self.address = address
            self.chksum_addr = to_checksum_address(value=address.hex_address)
            self.contract = self.contract_factory(self.chksum_addr)
            self.functions = self.contract.functions
        pass

    def constructor_data(self, *args: Any, **kwargs: Any) -> HexStr:
        data = self.contract_factory.constructor(*args, **kwargs).data_in_transaction
        return remove_0x_prefix( data )


    def call(self, cfn: ContractFunction) -> Any:
        """
        refactor ContractFunction.call() to this function
        :param cfn:
        :return:
        """
        call_transaction: TxParams = {}
        if 'data' in call_transaction:
            raise ValueError("Cannot set data in call transaction")

        if self.chksum_addr:
            call_transaction.setdefault('to', self.chksum_addr)
        # if self.web3.eth.defaultAccount is not empty:
        # type ignored b/c check prevents an empty defaultAccount
        # call_transaction.setdefault('from', self.web3.eth.defaultAccount)  # type: ignore

        pre_tx = prepare_transaction(
            self.chksum_addr,
            self.web3,
            fn_identifier=cfn.function_identifier,
            contract_abi=cfn.contract_abi,
            fn_abi=cfn.abi,
            transaction=call_transaction,
            fn_args=cfn.args,
            fn_kwargs=cfn.kwargs,
        )

        data = remove_0x_prefix(pre_tx['data'])

        ret_data = self.rpc.contract_call(contract_address=self.address.bech32_address(), hex_data=data)
        # print('ret_data is {}'.format(ret_data))

        ret_data_bytes = decode_hex(ret_data)

        if cfn.abi is None:
            fn_abi = find_matching_fn_abi(cfn.contract_abi, self.web3.codec, cfn.function_identifier, args, kwargs)

        output_types = get_abi_output_types(cfn.abi)

        try:
            output_data = self.web3.codec.decode_abi(output_types, ret_data_bytes)
        except DecodingError as e:
            # Provide a more helpful error message than the one provided by
            # eth-abi-utils
            is_missing_code_error = (
                    ret_data_bytes in ACCEPTABLE_EMPTY_STRINGS)
            # and self.web3.eth.getCode(address) in ACCEPTABLE_EMPTY_STRINGS)
            if is_missing_code_error:
                msg = (
                    "Could not transact with/call contract function, is contract "
                    "deployed correctly and chain synced?"
                )
            else:
                msg = (
                    "Could not decode contract function call {} return data {} for "
                    "output_types {}".format(
                        cfn.function_identifier,
                        ret_data_bytes,
                        output_types
                    )
                )
            raise BadFunctionCallOutput(msg) from e

        normalizers = tuple()
        _normalizers = itertools.chain(
            BASE_RETURN_NORMALIZERS,
            normalizers,
        )
        normalized_data = map_abi_data(_normalizers, output_types, output_data)

        if len(normalized_data) == 1:
            return normalized_data[0]
        else:
            return normalized_data

        # return call_transaction

    def __str__(self):
        pass



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
)

from eth_utils.toolz import (
    assoc,
    dissoc,
)

from htdfsdk.web3.types import (
    RPCEndpoint,
    RPCResponse,
)

if TYPE_CHECKING:
    from htdfsdk.web3 import Web3  # noqa: F401


def normalize_errors_middleware(
    make_request: Callable[[RPCEndpoint, Any], Any], web3: "Web3"
) -> Callable[[RPCEndpoint, Any], RPCResponse]:
    def middleware(method: RPCEndpoint, params: Any) -> RPCResponse:
        result = make_request(method, params)

        # As of v1.8, Geth returns errors when you request a
        # receipt for a transaction that is not in the chain.
        # It used to return a result of None, so we simulate the old behavior.

        if method == "eth_getTransactionReceipt" and "error" in result:
            is_geth = str(web3.clientVersion).startswith("Geth")
            if is_geth and result["error"]["code"] == -32000:
                return assoc(
                    dissoc(result, "error"),
                    "result",
                    None,
                )
            else:
                return result
        else:
            return result
    return middleware
